/* vxbAIOWDM.cpp - ACCES I/O Products card driver for VxWorks 7 */

/* Copyright (c) 2019 ACCES I/O Products, Inc.
 * 10623 Roselle Street San Diego California 92121-1506 USA
 * +1-858-550-9559 | +1-800-326-1649 | sales@accesio.com
 *
 * The right to copy, distribute, modify, or otherwise make use of this software
 * may be pursuant to the terms of an applicable ACCES I/O Products, Inc.
 * (ACCES) license agreement, or in the absence of such agreement, is limited to
 * use with ACCES hardware products.
 */

/*
 modification history
 --------------------
 12mar19,j2h  forked for vxbAIOWDM
 25feb19,j2h  forked for mPCIe-IDIO-8
 21dec18,j2h  Adding AIOWDM.dll equivalent API, changed kprintf to printf in aioShow, modified for PCIe-DIO-24HS
 13jun18,acc  Release 1.0.
 */

/*
 DESCRIPTION
 ===========
 This driver provides installation (probe/attach/show) and I/O routines for
 selected ACCES I/O Products I/O cards, including interrupt support. 

 This "Description" augments that found in the companion header file
 vxbaio.h. It is recommended that that file be read before continuing here.

 This driver is for use on a PCI-compatible bus. 

 Include Files
 -------------
 vxbAIOWDM.h


 User-Callable Functions
 -----------------------
 For details of user-callable functions, see comments with each. All return
 STATUS and if returning ERROR, set global errno. See each for details.

 aioClose( )            close the card, releasing it for another user
 aioIntCallbackInstall( ) set the user ISR callback function pointer
 aioIntClear( )         clear interrupts; normally done by the ISR
 aioIntDisable( )       disable interrupts
 aioIntEnable( )        enable interrupts; see it for other requirements
 aioIntPending( )       return TRUE if was generated but not cleared
 aioOpen( )             open the card for exclusive access
 aioShow( )             show the state of the driver and device
 aioGetDriverInfo()     retrieve the driver revision and build date/time

 Additional Functions for Internal Use
 -------------------------------------
 The following functions are not intended for direct use by an ordinary user.
 Calling them in a test program or from the VxWorks kernel shell may be of use
 to a driver developer. If ACCES_DEBUG is defined, LOCAL will resolve to empty
 and the functions will be visible globally. They are not included in any
 header file and must be IMPORTed explicitly.

 aioCloseForce( )       close the card with no error checks
 aioDebugLevelSet( )    set debug output verbosity level; 0 means none
 aioInt( )              service interrupts (the ISR)
 aioPciAttach( )        attach function for PCI bus
 aioPciDrvInit( )       initialize the driver
 aioPciProbe( )         probe function for PCI bus
 isEntryError( )           check entry conditions for user-callable functions
 isDriverOK( )             check that the driver appears to be installed

 Implementation Summary
 ======================
 **** CAUTION: VxWorks is very flexible and there are often several ways to code
 **** a feature. Consult ACCES I/O Products or VxWorks 7 documentation for
 **** alternatives. The approach taken here is based on VxWorks 7 release SR0530
 **** 2018-03-16.

 For an introduction to the VxWorks 7 I/O subsystem, drivers, and devices, see
 the "VxWorks(R) 7 Programmer's Guide," Edition 19, 29 March 2018, chapters 11
 "I/O System" and 12 "Devices", especially chapter 12 section "I/O Driver Model
 (335-355). The implementation notes in this file augment that section for this
 specific driver; reading that section first is recommended.

 Supporting a typical VxWorks VxBus-based device requires four high-level steps
 (further details under "Driver Definition and Installation" below):

 1. Only once, a driver for the type of device is declared statically with
 required data with a VXB_DRV_DEF macro. Top-level PCI bus setup code uses
 this data to call vxbDrvAdd() to register the driver with VxBus. That data
 typically includes pointers to the PCI "probe" and "attach" functions used in
 steps 2 and 3 next.

 2. During installation of a PCI bus, the top-level PCI bus setup code scans the
 bus and for each and every device found, calls the driver's "probe" function
 for every PCI driver registered in step 1. If a driver supports that device,
 it returns OK. (See vxbLib.c:vxbDevProbe() for "best match" refinements.)

 3. If the probe is successful, the PCI bus setup code then calls that driver's
 "attach" function. That function calls (directly or indirectly), vxbDevAdd()
 to add the device to the I/O subsystem, and performs other device-specific
 actions such as saving BAR* information, connecting and enabling interrupts,
 setting the baud rate for serial devices, etc.

 *BAR: Base Address Register. The virtual address of the device in "PCI
 Configuration Space"; see the Wikipedia source and its references given
 in the header file.

 4. Only once, the driver itself calls iosDrvInstall() to provide callback
 functions for user-callable, high-level open(), close(), read(), write(),
 ioctl(), create(), and remove() functions. These high-level functions are
 NOT supported by this driver, which has its own user-callable driver-specific
 functions. Nevertheless, iosDrvInstall() is called in this driver for two
 reasons:

 - The call to iosDrvInstall() passes NULL pointers for all of the callback
 functions so that user attempts to make the high-level calls return ERROR.

 - The devs command to the VxWorks kernel shell displays all devices and the
 driver number for each. We want ACCES cards to appear in that display and
 the call to iosDrvInstall() returns that driver number (it is the index to
 the driver callbacks table).

 The code to make this call to iosDrvInstall() may be part of:

 a) An initialization function in the driver called during VxWorks startup
 due to an INIT_RTN statement in the Component Description File (CDF) for
 the driver. However, this approach requires understanding the order in
 which initialization is carried out (because, for example, iosInit() must
 be called by VxWorks initialization code before any other IOS functions
 can be called, including iosDrvInstall()).
 b) The driver's Probe function after testing that that it has not already
 made the call. This is sub-optimal because the Probe function is called
 many more times than the Attach function.
 c) The driver's Attach function after testing that that it has not already
 made the call. This simple approach is taken here.

 Two critical data structures support the above:

 1. typedef struct vxbDev {...} VXB_DEV and
 typedef struct vxbDev* VXB_DEV_ID

 The vxbDev structure is allocated in step 2 above by the top-level PCI bus
 setup code and used to call the driver "probe" and "attach" functions. 
 
 It includes a member for the device's name (pName), a pointer through which
 the "probe" can send data to the "attach" (pVxbSoftc), a pointer to the driver
 data of step 1 above (pVxbDriver), and a pointer to the driver table entry
 where the callback function pointers of step 4 above are stored.
 
 It is defined in 
 [[--J2H "C:\WR0600\vxworks-7\pkgs_v2\os\drv\vxbus-2.0.0.0\core\src"]]
 $WIND_BASE/pkgs/os/drv/vxbus-<version>/core<-version>/src/vxbDevDef.h 

 2. The "device descriptor" structure. This all-important structure is described
 in detail below in section "aio Device Descriptor and Pointer to It". 
 
 For most devices supported by VxWorks, a device descriptor is dynamically 
 allocated for each instance of the device. Here we use a static array.


 Key implementation steps:
 - Top-level PCI bus enumeration code calls our aioPciProbe(), which returns
 OK when the device passed to it "matches" a device the driver supports.
 - As a result, for that device, the top-level PCI bus enumeration code calls our
 aioPciAttach() function. It calls iosDrvInstall() as described above,
 sets a module LOCAL pointer to our LOCAL device descriptor structure
 aioDD[n], gets and saves in the device descriptor the "handle" and BAR
 information necessary to access the device, sets up to receive interrupts for
 the device and finally, calls iosDevAdd() to add our device to the I/O
 subsystem.

 At that point, a user task can call our specific user-callable functions listed
 above to open/close, read/write, and control interrupts for the device.

 Additional major notes sections below are:
 - "Debugging": describes debugging features included by defining the ACCES_DEBUG
 macro.
 - "Local defines"
 - "AIO Device Descriptor and Pointer to It".
 - "Driver Definition and Installation": a general description of adding a driver
 to VxWorks 7, using this driver as an example.
 */

#include "vxbAIOWDM.h"

#include <arch/i86/vxbAccessArchLib.h>
#include <base/b_BOOL.h>
#include <errno.h>
#include <hwif/buslib/vxbPciLib.h>
#include <hwif/util/vxbAccess.h>
#include <hwif/util/vxbResourceLib.h>
#include <hwif/vxbus/vxbLib.h>
#include <hwif/vxbus/vxbMethod.h>
#include <ioLib.h>
#include <iosLib.h>
#include <private/kwriteLibP.h>
#include <stdio.h>
#include <subsys/int/vxbIntLib.h>
#include <taskLibCommon.h>
#include <types/vxTypesOld.h>
#include <types/vxWind.h>
#include <vxWorksCommon.h>

/*******************************************************************************
 Debugging
 ---------
 This section define macros and module-level variables to include and control
 debug code and messages. If ACCES_DEBUG is defined:

 - Define macro DBG_MSG (level, fmt, ...), which uses kprintf () to display a
 message if aioDebugLevel > level.

 - Define convenience macros DBG_ENTER, DBG_EXIT_OK, and DBG_MSG0--see below.

 - Define LOCAL (defined by VxWorks as static) as nothing so that all normally
 private variables and functions in this file are global and can be examined in
 a debugger or using the VxWorks kernel shell.

 [[--J2H fix: mustn't assume video card exists. Note: Original driver did this 
 more compatibly.]] 
 - In aioPciProbe(), sets _func_kwrite() (used by kprintf()) to kwriteVga().
 This enables use of DBG_MSG () in all functions in this driver without
 exception (kwriteVga() outputs to the PC Console with ~no initialization).

 - Define and initialize aioDebugLevel; valid values are:
 0: No debug output, even if compiled in because ACCES_DEBUG is defined.
 1: Minimal: message at each functions' entry and exit (with return value).
 2: Verbose: 1 + additional information during function execution.

 *****************************************************************************
 * CAUTION: If interrupts are enabled, generating debug output __may__ cause * 
 * the system to crash after a few interrupts. During testing, to avoid this,*
 * set aioDebugLevel to 0 or call aioDebugLevelSet(0) after boot        *
 * up (e.g., from the VxWorks kernel shell).                                 *
 *****************************************************************************

 - Define aio_pVxbDev. This is a local copy of the argument passed to our
 aioPciProbe() and aioPciAttach()) functions to enable it to be
 examined in the Workbench debugger or VxWorks kernel shell.
 */

#undef  ACCES_DEBUG
#define ACCES_DEBUG      // Comment out to remove debug code.

#ifdef ACCES_DEBUG
STATUS kwriteVga (char * pBuffer, size_t len);
int kprintf(const char * fmt, ... );

/* Make local (private) symbols global. */
#ifdef  LOCAL
#undef  LOCAL
#define LOCAL
#endif /* ACCES_DEBUG */
/* Variables to control debug; see details above. */
LOCAL int aioDebugLevel = 2; /* See valid values above. */
LOCAL VXB_DEV_ID aio_pVxbDev; /* So can examine in shell. */

/* DBG_MSG messages--includes function name and line. */
#define DBG_MSG(level, fmt, ...)                                                \
    do                                                                          \
        {                                                                       \
        if ( aioDebugLevel >= level )                                           \
            {                                                                   \
            if (_func_kprintf != NULL)                                          \
                {                                                               \
                (* _func_kprintf)                                               \
                        ("%s:%d: " fmt, __func__, __LINE__, ##__VA_ARGS__);     \
                }                                                               \
            }                                                                   \
        }                                                                       \
    while (0)

/* DBG_MSG0 messages--no function name and line. */
#define DBG_MSG0(level, fmt, ...)                                               \
    do                                                                          \
        {                                                                       \
        if ( aioDebugLevel >= level )                                           \
            {                                                                   \
            if (_func_kprintf != NULL)                                          \
                {                                                               \
                (* _func_kprintf)                                               \
                        (fmt, ##__VA_ARGS__);                                   \
                }                                                               \
            }                                                                   \
        }                                                                       \
    while (0)

void DBG_showRegisterAction(int level, TRegAct act, char * desc)
    {
    if ( aioDebugLevel >= level ) if (_func_kprintf != NULL)
    (* _func_kprintf) (eActionFmt[act.Action], desc, eActionNames[act.Action], act.BAR, act.offset, act.value, act.mask);
    }

/* Use DBG_ENTER as the first statement of every function. */
#define DBG_ENTER     DBG_MSG(1, "Enter %s.\n", __func__)

/* Use DBG_EXIT_OK as the last statement of a function before "return OK;". */
#define DBG_EXIT_OK   DBG_MSG(1, "Exit %s OK.\n", __func__)

#else   /* !ACCES_DEBUG */
#define DBG_MSG(level, fmt, ...)
#define DBG_MSG0(level, fmt, ...)
#define DBG_ENTER
#define DBG_EXIT_OK
#define DBG_showRegisterAction(x,y,z)
#endif  /* ACCES_DEBUG */

/*******************************************************************************
 * Local macros.
 */

/* Use EXIT_ERR as the last statement of a function before "return ERROR;"
 * for a simple error exit that only sets errno and reports it. */
#ifndef ACCES_DEBUG
#define EXIT_ERR(errnoVal)                      \
    errno = errnoVal
#else
#define EXIT_ERR(errnoVal)                      \
    do                                          \
        {                                       \
        errno = errnoVal;                       \
        DBG_MSG (1, "Exit ERROR %d\n.", errno); \
        }                                       \
    while (0)
#endif 
#define AIO_VENDORID  (0x494F)
#define AIO_DRIVER_REV (0.30)
#define AIO_DRIVER_REV_STR "0.30"
/*******************************************************************************
 * Local declarations.
 */
LOCAL int AIO_LastDevNum = -1;
/*
 AIO Device Descriptor and Pointer to It
 ---------------------------------------
 struct aioDD, type AIO_DD, pointer paioDD

 aioDD[] is an array of data structures that contains all required information to 
 use an ACCES card once it has been added to the I/O system.

 The first member is a DEV_HDR structure as is required for all VxBus devices.

 The first member of a DEV_HDR is a list pointer that links the device into the
 global device list (used, for example, by the shell devs command to display a
 list of devices).

 Because this version of this driver supports only up to eight cards in a system
 the aioDD[] struct is a module LOCAL variable in this file.

 Notes for individual members in the device descriptor:

 - taskId:
 To have exclusive access after calling aioOpen() until aioClose(),
 need way to "lock" the card for the current user (meaning return ERROR on
 calls from other tasks). Use the taskId to store the task ID of the calling
 task on aioOpen(). Only requests from that task will be permitted.

 - handle and bar:
 The handle is used to access the actual device registers, whether memory-
 mapped or in I/O register address space type (one member of the handle is
 the type--see pAdrRes below). The device registers are relative to a
 (virtual) base address, and that address is stored in *part* of a 32-bit
 "Base Address Register" (BAR). A device may use up to 6 BARs and a device
 driver and the device maker must agree on the use of the BARs.

 Thus accessing a register requires specifying three values: a handle,
 a BAR, and an offset. 

 FYI: VxWorks provides six functions to read and write a device register
 of 8, 16, or 32 bits: vxbRead8(), vxbWrite8(), ... vxbWrite32() (declared
 in vxbAccess.h; 64-bit versions are provided for some architectures). The
 first two parameters for all vxbReadXX/vxbWriteXX functions are the BAR
 and the offset.  (Contrast with sysInByte/sysOutByte which are I/O-specific
 and thus intel-architecture - specific.)

 - pAdrRes:
 A pointer to an "address resource" used in getting the "handle" to the
 card (the handle introduced just above.)

 A VxWorks Bus "resource" provides access to devices. There are three types
 of resource: 1) a memory-mapped device address, 2) a device address in I/O
 address space for systems with port "in" / "out" instructions (x86), and
 3) an interrupt. A resource virtualizes each of these so that a reads and
 writes to a device need not be concerned about the hardware details.

 To obtain a resource, the device driver allocates a resource of a particular
 type. For our cards several resource allocations are required (both done in our
 Attach function):

 pAdrRes = vxbResourceAlloc (pVxbDev, VXB_RES_IO, 2); // 2==bar#
 pIntRes = vxbResourceAlloc (pVxbDev, VXB_RES_IRQ, 0);

 vxbResourceAlloc() allocates a small memory block and sets it based on the
 resource type.

 The first argument for both points to the device node in the complete list
 of devices added to the system. It is either allocated by the user or, in
 for PCI devices, allocated by the VxBus code that enumerates the devices on
 the bus, and is then passed to the device driver probe and attach functions
 to identify the discovered device.

 The second argument is the resource type: VXB_RES_IO, VXB_RES_MEM, or
 VXB_RES_IRQ to specify the type of resource to allocate.

 The third argument identifies the resource to allocate depending on the
 type.
 - VXB_RES_IO and VXB_RES_MEM
 The desired Base Address Register (BAR); a PCI bus device has up to six
 BAR of either type, numbered 0-5. 
 - VXB_RES_IRQ: device-dependent, usually 0.

 Once the resource is allocated, for memory and I/O resources, the handle
 may be indirectly obtained from the resource--see the code.

 Technically, once the handle has been obtained, this address resource
 could be immediately freed. It is instead saved in our device descriptor
 for two reasons:
 - In device descriptors allocated dynamically, if the device is later
 deleted, the resource allocation must also be deleted.
 - To aid in debugging.

 -pIntRes:
 pIntRes points to the resource allocated to connect to the interrupts from
 the card by calling vxbIntConnect() (in our Attach function). See pAdrRes
 for general details. Unlike the address resource which could be freed after
 the handle is obtained, the interrupt resource must be for as long as
 interrupts from the device are to be enabled. Thus, it is saved in our
 device descriptor for that reason as well as the two reasons given for
 pAdrRes.
 */
struct AIO_DD
    {
	DEV_HDR devHdr; /* I/O device header; includes device
	 name and driver number */
	char * pProductName; /* ACCES product friendly name; set by aioAttach() to the name in the AIO_PCI_DEV_CFG structure (see "Match Table" below); example: "mPCIe-DIO-24S" */
	TASK_ID taskId; /* ID of task card was opened by; set by aioOpen(); NULL means closed.*/
	void * handle[6]; /* see "handle" above */
	ULONG bar[6]; /* BAR virtual address; see above */
	VXB_RESOURCE * pAdrRes; /* address resource described above */
	VXB_RESOURCE * pIntRes; /* interrupt resource described above */
	STATUS (*pIntCallback) (); /* for pointer to user function to call on an interrupt */
	BOOL intEnabled; /* TRUE if interrupts are enabled */
	int intCount; /* see aioInt() */

	TRegAct IRQ_Init; /* things like PCI Express cards needing BAR[0]+0x69 << 0x09 */
	TRegAct IRQ_Disable; /* enable all IRQ sources, or "primary" */
	TRegAct IRQ_Clear;
	TRegAct IRQ_Enable;
	TRegAct IRQ_IsLatched;

	void * IRQ_Generate; /* diagnostic: force card to generate an IRQ, [generally: if enabled] */
    };

/* Device descriptor and pointer:
 * Declare our module-level static device descriptor array and the pointer to it. The
 * pointer is set in aioPciAttach().
 */
LOCAL AIO_DD aioDD[AIO_MAX_DEVICES];

/*******************************************************************************
 Driver Definition and Installation
 ==================================
 Defining a driver and adding it to the I/O system requires several steps,
 including meeting consistency requirements described below. The steps are
 presented in top-down order here, but are implemented bottom-up to reduce
 the need for forward references.

 1. Declarations: Three declarations are required (illustrated using this ACCES
 driver):

 1.a. VXB_DRV_DEF (aioPciDrv)

 Add the driver to the system. Calls (directly or indirectly depending on
 the type of image):

 vxbDrvAdd (VXB_DRV * pDriver [= &aioPciDrv in 1.b]);

 vxbDrvAdd() registers the driver with the VxBus subsystem by linking the
 VxBus driver structure passed to it into the list of drivers rooted at
 global variable vxbDrvList (in vxbLib.c). vxbDrvAdd() then calls the Probe
 and Attach functions for any devices detected on the bus and having a
 matching class (as given by pDriver->vxbClass (VXB_BUSID_PCI here).

 1.b. LOCAL VXB_DRV aioPciDrv =
 {
 { NULL },               /@ SL_NODE          vxbNode: for drivers list @/
 "ACCES DIO PCI Driver", /@ char *           driver name @/
 "Driver for ACCES ...", /@ char *           description @/
 VXB_BUSID_PCI,          /@ vxb_BUSTYPE_ID   vxbClass        @/
 0,                      /@ UINT32           vxbFlags        @/
 0,                      /@ int              RefCnt          @/
 aioPciDrvMethods,       /@ VXB_DRV_METHOD * pVxbMethods--ptr to the
 methods table @/
 NULL                    /@ void *           pVxbParamDefaults @/
 };

 Declare and initialize a [[J2H]] static VXB_DRV structure as required by the
 VXB_DRV_DEF statement above. As noted, the structure will be linked into the
 global vxbDrvList. The structure contains a pointer to an array of callback
 functions specific to the driver--including generally for PCI devices the
 Probe and Attach functions--here aioPciProbe() and aioPciAttach()
 in this file.

 1.c. VXB_DRV_METHOD aioPciDrvMethods[] = { ... }

 Declare and initialize the "Methods Table" array of callback functions
 required by the VXB_DRV aioPciDrv structure described above.

 2. Executable statements

 Registering and installing a driver with the VxBus subsystem requires two
 function calls:

 - vxbDrvAdd (VXB_DRV * pDriver)
 As noted above, this is called indirectly as a result of the VXB_DRV_DEF()
 macro above.

 - iosDrvInstall() is called to set callbacks for seven basic I/O functions; as
 noted above, iosDrvInstall() is called in this driver but with all NULL
 pointers for the callback functions (see "Implementation Summary" step 4).

 At that point, devices for the driver may be added to the I/O subsystem,
 typically by calling a "create" or "add" function. For PCI devices, this is
 usually done in the Attach function (or a function it calls) as it is here.

 Consistency requirements
 ------------------------
 - Methods Table: The Methods Table member of the VXB_DRV structure (pVxbMethods)
 must point to the Methods Table in this file, here aioPciDrvMethods. See
 the VXB_DRV aioPciDrv structure below.
 */

/* Implementation of above, with reference paragraph numbers. */

/* 1.c Methods Table: Declare the table of VxBus PCI methods implemented in this
 * file. See above for purpose. See vxbMethod.h for macros VXB_DEVMETHOD_CALL
 * and VXB_DEVMETHOD_END. */

/* Forwards for our Attach and Probe functions required for the Methods table. */
LOCAL STATUS aioPciAttach (
			   VXB_DEV_ID pDev);
LOCAL STATUS aioPciProbe (
			  VXB_DEV_ID pDev);

LOCAL VXB_DRV_METHOD aioPciDrvMethods[] =
    {
	{
	VXB_DEVMETHOD_CALL(vxbDevProbe), aioPciProbe
	},
	{
	VXB_DEVMETHOD_CALL(vxbDevAttach), aioPciAttach
	},
    VXB_DEVMETHOD_END
    }
;

/* 1.b VxBus driver structure. See "Driver Definition and Installation" above for
 * details. */

VXB_DRV aioPciDrv =
    {
	{
	NULL
	}, /* SL_NODE          vxbNode: for drivers list */
    "ACCES Driver", /* char * */
    "Driver for ACCES I/O Products' PCI(e) bus cards",
    /* char *           description */
    VXB_BUSID_PCI, /* VXB_BUSTYPE_ID   vxbClass */
    0, /* UINT32           vxbFlags (see below) */
    0, /* int              RefCnt   */
    aioPciDrvMethods,
    /* VXB_DRV_METHOD * pVxbMethods; Methods Table ptr */
    };

/* vxbFlags: May be used to pass parameters to a driver as described in
 * pkgs/os/drv/vxbus-<version>/core-<version>/src/vxbParamLib.c. It is
 * invoked by providing a vxbFlags of VXB_DRVFLAG_PARAM instead of 0. When
 * used, the VXB_DRV structure has a final member pointing to a table of
 * parameter defaults. See vxbParamLib.c for details. */

/* 1.a Define the driver! Indirectly calls vxbDrvAdd(). See "Driver Definition
 * and Installation" for details. */
VXB_DRV_DEF(aioPciDrv);

/*
 * drvNum: Declare the module LOCAL driver number returned by iosDrvInstall()
 * called in aioPciAttach(). Special values:
 *  0:          Driver not yet initialized with the iosDrvInstall call.
 * -1 (ERROR):  Driver initialization, Probe, or Attach failed.
 */
LOCAL int drvNum = 0; /* Initialize to 0 to be explicit. */

/*
 * Match Table and Configuration Data
 * ----------------------------------
 * Match Table: The Match Table is normally an array of VXB_PCI_DEV_MATCH_ENTRY structs.
 * 
 * This driver uses a custom Match function instead of vxbPciDevMatch in order to
 * use a simpler structure: all supported cards share a single Vendor ID (0x494F), 
 * for example. 
 * 
 * There must be one unique entry in this table for each PCI device supported by
 * this driver as identified by the Vendor ID and Device ID of the device. The
 * table is searched by our Probe function each time it is called by the PCI
 * VxBus subsystem while enumerating the devices on the a PCI bus (see "Driver
 * Definition and Installation" above for more). A match means this driver
 * supports the device.
 *
 * The configuration structure type and instances are given first followed by
 * the Match Table.
 */

#define isLatched_PCI   {TESTBITS8, 	.BAR=1, .offset=0x4C, .value=0xFF, .mask=0x04 }
#define isLatched_PCIe  {TESTBITS8, 	.BAR=1, .offset=0x69, .value=0xFF, .mask=0x80 }
#define init_PCIe       {WRITE8,        .BAR=1, .offset=0x69, .value=0x09 }

/* IRQ Configuration struct type */
typedef struct
    {
	TRegAct Init; /* things like PCI Express cards needing BAR[0]+0x69 << 0x09 */
	TRegAct IsLatched; /* does this card have an IRQ pending? */
	TRegAct Disable; /* enable all IRQ sources, or "primary" */
	TRegAct Clear;
	TRegAct Enable;
	void * Generate; /*  diagnostic: force card to generate an IRQ, [generally: if enabled] */
    /*  TODO: add support for PCIe-IDIO-24-style IRQs
     *  TODO: add support for mPCIe-DIO-24A-style IRQs
     */
    } AIO_IRQ_STYLE;

/* IRQ Configuration data instances, one per different IRQ handling "style" used by supported devices. */
LOCAL AIO_IRQ_STYLE PCI_1_STYLE = /* S4C&04|13<00|13>|13<02	 */
    {
    .Init =
	{
	NotApplicable
	}, .IsLatched = isLatched_PCI,
    .Disable =
	{
	WRITE8, .BAR = 2, .offset = 0x13, .value = 0x00
	}, .Clear =
	{
	READ8, .BAR = 2, .offset = 0x13
	}, .Enable =
	{
	WRITE8, .BAR = 2, .offset = 0x13, .value = 0x02
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCI_2_STYLE = /* S4C&04|1E<||1F<	 */
    {
    .Init =
	{
	NotApplicable
	}, .IsLatched = isLatched_PCI,
    .Disable =
	{
	WRITE8, .BAR = 2, .offset = 0x1E
	}, .Clear =
	{
	NotApplicable
	}, .Enable =
	{
	WRITE8, .BAR = 2, .offset = 0x1F
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCI_3_STYLE = /* S4C&04|2<|1<|2>	 */
    {
    .Init =
	{
	NotApplicable
	}, .IsLatched = isLatched_PCI,
    .Disable =
	{
	WRITE8, .BAR = 2, .offset = 0x02
	}, .Clear =
	{
	WRITE8, .BAR = 2, .offset = 0x02
	}, .Enable =
	{
	READ8, .BAR = 2, .offset = 0x01
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCI_4_STYLE = /* S4C&04|4>|C>|3>	 */
    {
    .Init =
	{
	NotApplicable
	}, .IsLatched = isLatched_PCI,
    .Disable =
	{
	READ8, .BAR = 2, .offset = 0x04
	}, .Clear =
	{
	READ8, .BAR = 2, .offset = 0x0C
	}, .Enable =
	{
	READ8, .BAR = 2, .offset = 0x03
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCI_5_STYLE = /* S4C&04|7<00|7>|7<80	 */
    {
    .Init =
	{
	NotApplicable
	}, .IsLatched = isLatched_PCI,
    .Disable =
	{
	WRITE8, .BAR = 2, .offset = 0x07, .value = 0x00
	}, .Clear =
	{
	READ8, .BAR = 2, .offset = 0x07
	}, .Enable =
	{
	WRITE8, .BAR = 2, .offset = 0x07, .value = 0x80
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCI_6_STYLE = /* S4C&04|9>||4<81	 */
    {
    .Init =
	{
	NotApplicable
	}, .IsLatched = isLatched_PCI,
    .Disable =
	{
	READ8, .BAR = 2, .offset = 0x09
	}, .Clear =
	{
	NotApplicable
	}, .Enable =
	{
	WRITE8, .BAR = 2, .offset = 0x04, .value = 0x81
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCI_7_STYLE = /* S4C&04|9>||4>	 */
    {
    .Init =
	{
	NotApplicable
	}, .IsLatched = isLatched_PCI,
    .Disable =
	{
	READ8, .BAR = 2, .offset = 0x09
	}, .Clear =
	{
	NotApplicable
	}, .Enable =
	{
	READ8, .BAR = 2, .offset = 0x04
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCI_8_STYLE = /* S4C&04|B<FF|F<|B<00	 */
    {
    .Init =
	{
	NotApplicable
	}, .IsLatched = isLatched_PCI,
    .Disable =
	{
	WRITE8, .BAR = 2, .offset = 0x0B, .value = 0xFF
	}, .Clear =
	{
	WRITE8, .BAR = 2, .offset = 0x0F
	}, .Enable =
	{
	WRITE8, .BAR = 2, .offset = 0x0B, .value = 0x00
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCI_9_STYLE = /* S4C&04|C<00||C<10	 */
    {
    .Init =
	{
	NotApplicable
	}, .IsLatched = isLatched_PCI,
    .Disable =
	{
	WRITE8, .BAR = 2, .offset = 0x0C, .value = 0x00
	}, .Clear =
	{
	NotApplicable
	}, .Enable =
	{
	WRITE8, .BAR = 2, .offset = 0x0C, .value = 0x10
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCI_10_STYLE = /* S4C&04|D<|F<|E<	 */
    {
    .Init =
	{
	NotApplicable
	}, .IsLatched = isLatched_PCI,
    .Disable =
	{
	WRITE8, .BAR = 2, .offset = 0x0D
	}, .Clear =
	{
	WRITE8, .BAR = 2, .offset = 0x0F
	}, .Enable =
	{
	WRITE8, .BAR = 2, .offset = 0x0E
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCI_11_STYLE = /* S4C&04|E<00|8#6|E<FF	 */
    {
    .Init =
	{
	NotApplicable
	}, .IsLatched = isLatched_PCI,
    .Disable =
	{
	WRITE8, .BAR = 2, .offset = 0x0E, .value = 0x00
	}, .Clear =
	{
	NOT_YET_IMPLEMENTED
	}, .Enable =
	{
	WRITE8, .BAR = 2, .offset = 0x0E, .value = 0xFF
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCI_12_STYLE = /* S4C&04||4>|	 	 */
    {
    .Init =
	{
	NotApplicable
	}, .IsLatched = isLatched_PCI,
    .Disable =
	{
	NotApplicable
	}, .Clear =
	{
	READ8, .BAR = 2, .offset = 0x04
	}, .Enable =
	{
	NotApplicable
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCI_13_STYLE = /* S4C&04||7F<00|	 */
    {
    .Init =
	{
	NotApplicable
	}, .IsLatched = isLatched_PCI,
    .Disable =
	{
	NotApplicable
	}, .Clear =
	{
	WRITE8, .BAR = 2, .offset = 0x7F, .value = 0x00
	}, .Enable =
	{
	NotApplicable
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCI_NO_IRQ_STYLE = /* S4C&04|||	 	 */
    {
    .Init =
	{
	NotApplicable
	}, .IsLatched = isLatched_PCI,
    .Disable =
	{
	NotApplicable
	}, .Clear =
	{
	NotApplicable
	}, .Enable =
	{
	NotApplicable
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCIE_1_STYLE = /* S69&80|1E<||1F<	 */
    {
    .Init = init_PCIe,
    .IsLatched = isLatched_PCIe,
    .Disable =
	{
	WRITE8, .BAR = 2, .offset = 0x1E
	}, .Clear =
	{
	NotApplicable
	}, .Enable =
	{
	WRITE8, .BAR = 2, .offset = 0x1F
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCIE_2_STYLE = /* S69&80|28<00|29<FF|28<FF	 */
    {
    .Init = init_PCIe,
    .IsLatched = isLatched_PCIe,
    .Disable =
	{
	WRITE8, .BAR = 2, .offset = 0x28, .value = 0x00
	}, .Clear =
	{
	WRITE8, .BAR = 2, .offset = 0x29, .value = 0xFF
	}, .Enable =
	{
	WRITE8, .BAR = 2, .offset = 0x28, .value = 0xFF
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCIE_3_STYLE = /* S69&80|2<|1<|2>	 */
    {
    .Init = init_PCIe,
    .IsLatched = isLatched_PCIe,
    .Disable =
	{
	WRITE8, .BAR = 2, .offset = 0x02
	}, .Clear =
	{
	WRITE8, .BAR = 2, .offset = 0x01
	}, .Enable =
	{
	READ8, .BAR = 2, .offset = 0x02
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCIE_4_STYLE = /* S69&80|B<FF|F<|B<00	 */
    {
    .Init = init_PCIe,
    .IsLatched = isLatched_PCIe,
    .Disable =
	{
	WRITE8, .BAR = 2, .offset = 0x0B, .value = 0xFF
	}, .Clear =
	{
	WRITE8, .BAR = 2, .offset = 0x0F
	}, .Enable =
	{
	WRITE8, .BAR = 2, .offset = 0x0B, .value = 0x00
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCIE_5_STYLE = /* S69&80|B>|F<|B<00	 */
    {
    .Init = init_PCIe,
    .IsLatched = isLatched_PCIe,
    .Disable =
	{
	READ8, .BAR = 2, .offset = 0x0B
	}, .Clear =
	{
	WRITE8, .BAR = 2, .offset = 0x0F
	}, .Enable =
	{
	WRITE8, .BAR = 2, .offset = 0x0B, .value = 0x00
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCIE_6_STYLE = /* S69&80|E<00|8#4|E<FF	 */
    {
    .Init = init_PCIe,
    .IsLatched = isLatched_PCIe,
    .Disable =
	{
	WRITE8, .BAR = 2, .offset = 0x0E, .value = 0x00
	}, .Clear =
	{
	WRITE32, .BAR = 2, .offset = 0x08, .value = 0xFFFFFFFF
	}, .Enable =
	{
	WRITE8, .BAR = 2, .offset = 0x0E, .value = 0xFF
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCIE_7_STYLE = /* S69&80|E<00|F<|E<03	 */
    {
    .Init = init_PCIe,
    .IsLatched = isLatched_PCIe,
    .Disable =
	{
	WRITE8, .BAR = 2, .offset = 0x0E, .value = 0x00
	}, .Clear =
	{
	WRITE8, .BAR = 2, .offset = 0x0F
	}, .Enable =
	{
	WRITE8, .BAR = 2, .offset = 0x0E, .value = 0x03
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE PCIE_8_STYLE = /* S4C&04|4>|C>|3>	 */
    {
    .Init = init_PCIe,
    .IsLatched = isLatched_PCIe,
    .Disable =
	{
	READ8, .BAR = 2, .offset = 0x04
	}, .Clear =
	{
	READ8, .BAR = 2, .offset = 0x0C
	}, .Enable =
	{
	READ8, .BAR = 2, .offset = 0x03
	}, .Generate = NULL,
    };

LOCAL AIO_IRQ_STYLE NYI =
    {
    .Init =
	{
	NotApplicable
	}, .IsLatched =
	{
	NotApplicable
	}, .Disable =
	{
	NotApplicable
	}, .Clear =
	{
	NotApplicable
	}, .Enable =
	{
	NotApplicable
	}, .Generate = NULL,
    };

typedef struct /* Configuration data structure type. */
    {
	UINT16 pciDevId; /* PCI Device ID */
	char * Model; /* Model name */
	AIO_IRQ_STYLE * IRQ; /* IRQ control/configuration data, not an IRQ Number */
	UINT32 flags; /* bitmasked flag storage */
    } AIO_PCI_DEV_CFG;

/* Match Table */
LOCAL const AIO_PCI_DEV_CFG matchTable[] =
    {
	{
	.pciDevId = 0x0100, .Model = "mPCIe-DIO-24S", .IRQ = &PCIE_2_STYLE,
	},
	{
	.pciDevId = 0x0101, .Model = "mPCIe-IDIO-8", .IRQ = &PCIE_2_STYLE,
	},
	{
	.pciDevId = 0x0102, .Model = "mPCIe-IIRO-8", .IRQ = &PCIE_2_STYLE,
	},
	{
	.pciDevId = 0x0103, .Model = "mPCIe-IDIO-4", .IRQ = &PCIE_2_STYLE,
	},
	{
	.pciDevId = 0x0104, .Model = "mPCIe-IIRO-4", .IRQ = &PCIE_2_STYLE,
	},
	{
	.pciDevId = 0x0105, .Model = "mPCIe-IDO-8", .IRQ = &PCIE_2_STYLE,
	},
	{
	.pciDevId = 0x0106, .Model = "mPCIe-RO-8", .IRQ = &PCIE_2_STYLE,
	},
	{
	.pciDevId = 0x0107, .Model = "mPCIe-II-16", .IRQ = &PCIE_2_STYLE,
	},
	{
	.pciDevId = 0x0108, .Model = "mPCIe-II-8", .IRQ = &PCIE_2_STYLE,
	},
	{
	.pciDevId = 0x0109, .Model = "mPCIe-II-4", .IRQ = &PCIE_2_STYLE,
	},
	{
	.pciDevId = 0x010A, .Model = "mPCIe-QUAD-4", .IRQ = &PCIE_2_STYLE,
	},
	{
	.pciDevId = 0x010B, .Model = "mPCIe-QUAD-8", .IRQ = &PCIE_2_STYLE,
	},
	{
	.pciDevId = 0x010C, .Model = "mod-TBD", .IRQ = &NYI,
	},
	{
	.pciDevId = 0x010D, .Model = "mod-TBD", .IRQ = &NYI,
	},
	{
	.pciDevId = 0x010E, .Model = "mod-TBD", .IRQ = &NYI,
	},
	{
	.pciDevId = 0x010F, .Model = "mod-TBD", .IRQ = &NYI,
	},
	{
	.pciDevId = 0x0508, .Model = "PCI-IDO-16A", .IRQ = &PCI_NO_IRQ_STYLE,
	},
	{
	.pciDevId = 0x0518, .Model = "PCI-IDO-32A", .IRQ = &PCI_NO_IRQ_STYLE,
	},
	{
	.pciDevId = 0x0520, .Model = "PCI-IDO-48", .IRQ = &PCI_NO_IRQ_STYLE,
	},
	{
	.pciDevId = 0x0521, .Model = "PCI-IDO-48A", .IRQ = &PCI_NO_IRQ_STYLE,
	},
	{
	.pciDevId = 0x0703, .Model = "PCIe-RO-4", .IRQ = &PCIE_3_STYLE,
	},
	{
	.pciDevId = 0x07C0, .Model = "PCIe-IDO-12", .IRQ = &NYI,
	},
	{
	.pciDevId = 0x07D0, .Model = "PCIe-IDO-24", .IRQ = &PCIE_6_STYLE,
	},
	{
	.pciDevId = 0x0920, .Model = "PCI-IDI-48", .IRQ = &PCI_5_STYLE,
	},
	{
	.pciDevId = 0x0BC0, .Model = "PCIe-IDI-12A", .IRQ = &NYI,
	},
	{
	.pciDevId = 0x0BC1, .Model = "PCIe-IDI-12B", .IRQ = &NYI,
	},
	{
	.pciDevId = 0x0BD0, .Model = "PCIe-IDI-24", .IRQ = &PCIE_6_STYLE,
	},
	{
	.pciDevId = 0x0BD1, .Model = "PCIe-IDI-24B", .IRQ = &NYI,
	},
	{
	.pciDevId = 0x0C40, .Model = "PCIe-DIO-144", .IRQ = &PCIE_1_STYLE,
	},
	{
	.pciDevId = 0x0C50, .Model = "PCI-DIO-24H", .IRQ = &PCI_10_STYLE,
	},
	{
	.pciDevId = 0x0C51, .Model = "PCI-DIO-24D", .IRQ = &PCI_10_STYLE,
	},
	{
	.pciDevId = 0x0C52, .Model = "PCIe-DIO-24", .IRQ = &PCIE_7_STYLE,
	},
	{
	.pciDevId = 0x0C53, .Model = "PCIe-DIO-24D/H", .IRQ = &PCIE_7_STYLE,
	},
	{
	.pciDevId = 0x0C54, .Model = "PCIe-DIO-24HC", .IRQ = &PCIE_4_STYLE,
	},
	{
	.pciDevId = 0x0C57, .Model = "mPCIe-DIO-24", .IRQ = &PCIE_4_STYLE,
	},
	{
	.pciDevId = 0x0C60, .Model = "PCI-DIO-48(H)", .IRQ = &PCI_8_STYLE,
	},
	{
	.pciDevId = 0x0C61, .Model = "PCIe-DIO-48", .IRQ = &PCIE_4_STYLE,
	},
	{
	.pciDevId = 0x0C62, .Model = "P104-DIO-48", .IRQ = &PCI_NO_IRQ_STYLE,
	},
	{
	.pciDevId = 0x0C68, .Model = "PCI-DIO-72", .IRQ = &PCI_2_STYLE,
	},
	{
	.pciDevId = 0x0C69, .Model = "P104-DIO-96", .IRQ = &PCI_NO_IRQ_STYLE,
	},
	{
	.pciDevId = 0x0C6A, .Model = "PCIe-DIO-72", .IRQ = &PCIE_1_STYLE,
	},
	{
	.pciDevId = 0x0C70, .Model = "PCI-DIO-96", .IRQ = &PCI_2_STYLE,
	},
	{
	.pciDevId = 0x0C71, .Model = "PCIe-DIO-96", .IRQ = &PCIE_1_STYLE,
	},
	{
	.pciDevId = 0x0C78, .Model = "PCI-DIO-120", .IRQ = &PCI_2_STYLE,
	},
	{
	.pciDevId = 0x0C79, .Model = "PCIe-DIO-120", .IRQ = &PCIE_1_STYLE,
	},
	{
	.pciDevId = 0x0DC8, .Model = "PCI-IDIO-16", .IRQ = &PCI_3_STYLE,
	},
	{
	.pciDevId = 0x0E50, .Model = "PCI-DIO-24S", .IRQ = &PCI_8_STYLE,
	},
	{
	.pciDevId = 0x0E51, .Model = "PCI-DIO-24H(C)", .IRQ = &PCI_10_STYLE,
	},
	{
	.pciDevId = 0x0E52, .Model = "PCI-DIO-24D(C)", .IRQ = &PCI_10_STYLE,
	},
	{
	.pciDevId = 0x0E53, .Model = "PCIe-DIO-24S", .IRQ = &PCIE_4_STYLE,
	},
	{
	.pciDevId = 0x0E54, .Model = "PCIe-DIO-24DS/HS", .IRQ = &PCIE_4_STYLE,
	},
	{
	.pciDevId = 0x0E55, .Model = "PCIe-DIO-24DC", .IRQ = &PCIE_7_STYLE,
	},
	{
	.pciDevId = 0x0E56, .Model = "PCIe-DIO-24DCS", .IRQ = &PCIE_4_STYLE,
	},
	{
	.pciDevId = 0x0E57, .Model = "PCI-DIO-48HS", .IRQ = &PCIE_4_STYLE,
	},
	{
	.pciDevId = 0x0E60, .Model = "PCI-DIO-48S", .IRQ = &PCI_8_STYLE,
	},
	{
	.pciDevId = 0x0E61, .Model = "PCIe-DIO-48S", .IRQ = &PCIE_4_STYLE,
	},
	{
	.pciDevId = 0x0E62, .Model = "P104-DIO-48S", .IRQ = &PCI_11_STYLE,
	},
	{
	.pciDevId = 0x0E63, .Model = "PCIe-EZ-BLEND", .IRQ = &NYI,
	},
	{
	.pciDevId = 0x0E6A, .Model = "PCIe-DIO-72S", .IRQ = &PCIE_1_STYLE,
	},
	{
	.pciDevId = 0x0E71, .Model = "PCIe-DIO-96S", .IRQ = &PCIE_1_STYLE,
	},
	{
	.pciDevId = 0x0E79, .Model = "PCIe-DIO-120S", .IRQ = &PCIE_1_STYLE,
	},
	{
	.pciDevId = 0x0F00, .Model = "PCI-IIRO-8", .IRQ = &PCI_3_STYLE,
	},
	{
	.pciDevId = 0x0F01, .Model = "LPCI-IIRO-8", .IRQ = &PCI_3_STYLE,
	},
	{
	.pciDevId = 0x0F02, .Model = "PCIe-IIRO-8", .IRQ = &PCIE_3_STYLE,
	},
	{
	.pciDevId = 0x0F08, .Model = "PCI-IIRO-16", .IRQ = &PCI_3_STYLE,
	},
	{
	.pciDevId = 0x0F09, .Model = "PCIe-IIRO-16", .IRQ = &PCIE_3_STYLE,
	},
	{
	.pciDevId = 0x0FC0, .Model = "PCIe-IDIO-12", .IRQ = &PCIE_6_STYLE,
	},
	{
	.pciDevId = 0x0FC1, .Model = "PCIe-IDI-12", .IRQ = &PCIE_6_STYLE,
	},
	{
	.pciDevId = 0x0FC2, .Model = "PCIe-IDO-12", .IRQ = &PCIE_6_STYLE,
	},
	{
	.pciDevId = 0x0FD0, .Model = "PCIe-IDIO-24", .IRQ = &PCIE_6_STYLE,
	},
	{
	.pciDevId = 0x0FD1, .Model = "PCIe-IDIO-24B", .IRQ = &NYI,
	},
	{
	.pciDevId = 0x2230, .Model = "PCI-QUAD-8", .IRQ = &PCI_13_STYLE,
	},
	{
	.pciDevId = 0x2231, .Model = "PCI-QUAD-4", .IRQ = &PCI_13_STYLE,
	},
	{
	.pciDevId = 0x22C0, .Model = "PCI-WDG-CSM", .IRQ = &PCI_7_STYLE,
	},
	{
	.pciDevId = 0x25C0, .Model = "P104-WDG-E", .IRQ = &PCI_6_STYLE,
	},
	{
	.pciDevId = 0x2C50, .Model = "PCI-DIO-96CT", .IRQ = &PCI_2_STYLE,
	},
	{
	.pciDevId = 0x2C58, .Model = "PCI-DIO-96C3", .IRQ = &PCI_2_STYLE,
	},
	{
	.pciDevId = 0x2C59, .Model = "PCIe-DIO-96C3", .IRQ = &PCIE_1_STYLE,
	},
	{
	.pciDevId = 0x2C70, .Model = "PCI-DIO-96CT", .IRQ = &NYI,
	},
	{
	.pciDevId = 0x2E50, .Model = "mPCIe-DIO-24A", .IRQ = &PCIE_4_STYLE,
	},
	{
	.pciDevId = 0x2EE0, .Model = "PCIe-DIO24S-CTR12", .IRQ = &PCIE_5_STYLE,
	},
	{
	.pciDevId = 0x2FC0, .Model = "P104-WDG-CSM", .IRQ = &PCI_6_STYLE,
	},
	{
	.pciDevId = 0x2FC1, .Model = "P104-WDG-CSMA", .IRQ = &PCI_6_STYLE,
	},
	{
	.pciDevId = 0x4890, .Model = "PCIe-DA12-2", .IRQ = &PCIE_8_STYLE,
	},
	{
	.pciDevId = 0x4898, .Model = "PCIe-DA12-4", .IRQ = &PCIE_8_STYLE,
	},
	{
	.pciDevId = 0x48a0, .Model = "PCIe-DA12-6", .IRQ = &PCIE_8_STYLE,
	},
	{
	.pciDevId = 0x48a8, .Model = "PCIe-DA12-8", .IRQ = &PCIE_8_STYLE,
	},
	{
	.pciDevId = 0x48b0, .Model = "PCIe-DA12-16", .IRQ = &PCIE_8_STYLE,
	},
	{
	.pciDevId = 0x48d0, .Model = "PCIe-DA16-2", .IRQ = &PCIE_8_STYLE,
	},
	{
	.pciDevId = 0x48d8, .Model = "PCIe-DA16-4", .IRQ = &PCIE_8_STYLE,
	},
	{
	.pciDevId = 0x48e0, .Model = "PCIe-DA16-6", .IRQ = &PCIE_8_STYLE,
	},
	{
	.pciDevId = 0x48e8, .Model = "PCIe-DA16-8", .IRQ = &PCIE_8_STYLE,
	},
	{
	.pciDevId = 0x48f0, .Model = "PCIe-DA16-16", .IRQ = &PCIE_8_STYLE,
	},
	{
	.pciDevId = 0x5ED0, .Model = "PCI-DAC", .IRQ = &NYI,
	},
	{
	.pciDevId = 0x6C90, .Model = "PCI-DA12-2", .IRQ = &PCI_4_STYLE,
	},
	{
	.pciDevId = 0x6C98, .Model = "PCI-DA12-4", .IRQ = &PCI_4_STYLE,
	},
	{
	.pciDevId = 0x6CA0, .Model = "PCI-DA12-6", .IRQ = &PCI_4_STYLE,
	},
	{
	.pciDevId = 0x6CA8, .Model = "PCI-DA12-8", .IRQ = &PCI_4_STYLE,
	},
	{
	.pciDevId = 0x6CA9, .Model = "PCI-DA12-8V", .IRQ = &PCI_4_STYLE,
	},
	{
	.pciDevId = 0x6CB0, .Model = "PCI-DA12-16", .IRQ = &PCI_4_STYLE,
	},
	{
	.pciDevId = 0x6CB1, .Model = "PCI-DA12-16V", .IRQ = &PCI_4_STYLE,
	},
	{
	.pciDevId = 0xACA8, .Model = "PCI-AI12-16", .IRQ = &PCI_12_STYLE,
	},
	{
	.pciDevId = 0xACA9, .Model = "PCI-AI12-16A", .IRQ = &PCI_12_STYLE,
	},
	{
	.pciDevId = 0xECA8, .Model = "PCI-AIO12-16", .IRQ = &PCI_12_STYLE,
	},
	{
	.pciDevId = 0xECA9, .Model = "PCI-A12-16", .IRQ = &PCI_12_STYLE,
	},
	{
	.pciDevId = 0xECAA, .Model = "PCI-A12-16A", .IRQ = &PCI_12_STYLE,
	},
	{
	.pciDevId = 0xECE8, .Model = "LPCI-A16-16A", .IRQ = &PCI_9_STYLE,
	},
	{
	.pciDevId = 0xECE9, .Model = "LPCI-AIO16A", .IRQ = &PCI_1_STYLE,
	},
	{
	} /* Empty terminates list */
    };

/* Forwards (continued) */
void aioCloseForce (
		    AIO_DD * dev);
STATUS aioIntCallbackInstall (
			      AIO_DD * dev,
			      STATUS (*callback_funcptr) ());
LOCAL BOOL isDriverOK (
		       AIO_DD *dev);
LOCAL BOOL isEntryError (
			 AIO_DD * dev);
UINT32 aioDoRegisterAction (
			 AIO_DD * dev,
			 TRegAct act,
			 char * desc);
STATUS aioIntDisable (
		      AIO_DD * dev);

/* Functions =================================================================*/

/*******************************************************************************
 *
 * aioClose - close the DIO card and make it available to another user
 *
 * This function closes the card and makes it available to another task. It is
 * not an error to close the card if it is not open. It is an error for a task
 * to attempt to close an open the card if it is not the owner (that is, if it is
 * not the task that opened it). Closing the card disables interrupts if enabled,
 * and always clears any pending interrupts and resets the ISR callback function
 * pointer to NULL.
 *
 * RETURNS: OK or ERROR per errno.
 *
 * ERRNO
 *   ENXIO       Driver not installed.
 *   EACCES      Card is open for another task.
 */

STATUS aioClose (
		 AIO_DD *dev)
    {
    DBG_ENTER;

    if (!isDriverOK (dev))
	{
	EXIT_ERR(ENXIO);
	return (ERROR);
	}

    if (dev->taskId != taskIdSelf ()) /* Another owner? */
	{
	EXIT_ERR(EACCES);
	return (ERROR);
	}

    /* With error checks passed, use the "force" version of this function
     * instead of repeating its code. */

    aioCloseForce (dev);

    DBG_EXIT_OK;
    return (OK);
    }

/*******************************************************************************
 *
 * aioCloseForce - close the DIO card even if not the owner
 *
 * This function closes the card even if the calling task is not the owner. No
 * error checks. It is intended for internal use and is not in the header file.
 * See aioClose() for the user-callable "safe" version.
 *
 * RETURNS: N/A
 */

void aioCloseForce (
		    AIO_DD *dev)
    {
    DBG_ENTER;

    /* First disable interrupts if enabled (which clears any pending interrupts),
     * and reset the ISR callback function pointer to NULL. Ignore any error from
     * aioIntDisable() since this is the "force" function. */

    aioIntDisable (dev);
    errno = 0; /* In case aioIntDisble() failed. */
    dev->pIntCallback = NULL;

    /* "Close" the card by setting the taskId of the current "owner" (if any) to
     * NULL. (Do this last so other functions called in this function won't
     * fail.) */

    dev->taskId = 0;

    DBG_EXIT_OK;
    }

#ifdef ACCES_DEBUG
/*******************************************************************************
 *
 * aioDebugLevelSet - set verbosity for debug output
 *
 * Set aioDebugLevel to control debug output verbosity. See "Debugging"
 * topic above for valid values.
 *
 * RETURNS: N/A
 */

void aioDebugLevelSet (int verbosityLevel)
    {
    aioDebugLevel = verbosityLevel;
    }

#endif /* ACCES_DEBUG */

/******************************************************************************
 *
 * aioGetDriverRevision - show driver revision and build date/time
 *
 * Show the current driver revision and build date/time. Does not require 
 * hardware. Does not generate errors.
 *
 * RETURNS: revision as a double, intended as ("1.00")
 *
 * ERRNO: N/A
 */
double aioGetDriverRevision (
			     void)
    {
    printf ("vxbAIOWDM Driver Revision %s %s %s\n", (AIO_DRIVER_REV_STR),
    __DATE__,
	    __TIME__);
    return (AIO_DRIVER_REV);
    }

/*******************************************************************************
 *
 * aioInt - interrupt service routine to handle an interrupt
 *
 * Handle an interrupt from the ACCES DIO card. Call the user callback function
 * if:
 * - Interrupts are enabled (by a prior call to aioIntEnable()).
 * - The current value of the ISR callback function pointer is not NULL. The
 *   callback ISR pointer is set by calling aioIntCallbackInstall().
 * - The card is open. (NOTE: the task ID is NOT checked--any task may be running
 *   when an interrupt occurs, not just the task that opened the card).
 *
 * Whether a user callback function is called or not, always clear the interrupt
 * at the end of this ISR (otherwise we would immediately receive another
 * interrupt).
 *
 * RETURNS: If driver initialization, probe, or attach failed, return ERROR per
 * errno. If the card is not open, interrupts are disabled, or there is no
 * non-NULL callback function pointer, return OK. Else return the value returned
 * by the callback function. (NOTE: Error ENXIO returned by the ISR cannot be
 * distinguished from ENXIO returned by the driver check in this function.)
 *
 * ERRNO
 *   ENXIO       Driver not installed.
 *
 * *****************************************************************************
 * * CAUTION: If interrupts are enabled, generating debug output may cause the *
 * * system to crash after a few interrupts. See this caution above under the  *
 * * "Debugging" topic for how to avoid this.
 * *****************************************************************************
 *
 * Note: The argument passed to this ISR when an interrupt occurs is set up by
 * the call to vxbIntConnect() in our Attach function--it is the fourth
 * argument passed to vxbIntConnect(). 
 * 
 */

LOCAL STATUS aioInt (
		     AIO_DD * dev)
    {
    STATUS status;

    dev->intCount++; /* For debug but generally useful. */

    DBG_ENTER;

    if (dev == NULL)
	{
	EXIT_ERR(ENXIO);
	return (ERROR);
	}

    if (aioDoRegisterAction (dev, dev->IRQ_IsLatched, "IRQ_IsLatched"))
	if (dev->intEnabled && (dev->pIntCallback != NULL)
	&& (dev->taskId != 0))
	    {
	    status = (dev->pIntCallback) ();

	    DBG_MSG (1, "was us, callback occurred, Exit. status=%i\n", status);
	    }
	else
	    {
	    status = OK;
	    DBG_MSG (1, "Exit. No user callback function installed.\n");
	    }
    else
	{
	status = ERROR;
	DBG_MSG(2,"Exit. Not our IRQ \n");
	}

    return (status);
    }

/*******************************************************************************
 *
 * aioIntCallbackInstall - install the user interrupt service routine (ISR)
 *
 * Record the callback interrupt service routine (ISR) function pointer argument.
 * The prototype for that function must be:
 *
 *   STATUS <ISR function name> (void)
 *
 * The device must be open by calling aioOpen() before calling this routine.
 *
 * The ISR will be called if:
 * - This routine has been called to establish the ISR callback function pointer.
 * - Interrupts are enabled by calling aioIntEnable().
 * - There is a change of state (CoS) on a pin on a port configured for input
 *   while interrupts are enabled.
 *
 * CoS will be ignored if interrupts are not then enabled, and are ignored for
 * pins on a port set for output direction.
 *
 * This routine may be called at any time that the card is open to change the
 * callback function pointer. It may be called with NULL function pointer to
 * remove the callback function.
 *
 * RETURNS: OK or ERROR per errno.
 *
 * ERRNO:
 *   ENXIO       Driver not installed.
 *   ENODEV      Card is not open.
 *   EACCES      Card is open for another task.
 */

STATUS aioIntCallbackInstall (
			      AIO_DD * dev,
			      STATUS (*callback_funcptr) (
							  void) /* callback function pointer */
			      )
    {
    DBG_MSG (1, "Enter. callback_funcptr=%p\n", callback_funcptr);

    if (isEntryError (dev))
	return (ERROR);

    dev->pIntCallback = callback_funcptr;

    DBG_EXIT_OK;
    return (OK);
    }

/*******************************************************************************
 *
 * aioIntClear - clear a pending interrupt, if any
 *
 * If an interrupt is pending, this function clears it; otherwise there is no
 * effect. NOTE: It is unusual for a user task to call this function because
 * interrupts are cleared as part of the interrupt service routine.
 *
 * RETURNS: OK or ERROR per errno.
 *
 * ERRNO
 *   ENXIO       Driver not installed.
 *   ENODEV      Card is not open.
 *   EACCES      Card is open for another task.
 */

STATUS aioIntClear (
		    AIO_DD * dev)
    {
    DBG_ENTER;

    if (isEntryError (dev))
	return (ERROR);

    aioDoRegisterAction (dev, dev->IRQ_Clear, "IRQ_Clear");

    DBG_EXIT_OK;
    return (OK);
    }

/*******************************************************************************
 *
 * aioIntDisable - disable interrupts from the card
 *
 * This routine disables interrupts from the card. Any pending interrupts are
 * cleared. Calling this routine when interrupts are disabled is not an error.
 *
 * RETURNS: OK or ERROR per errno.
 *
 * ERRNO
 *   ENXIO       Driver not installed.
 *   ENODEV      Card is not open.
 *   EACCES      Card is open for another task.
 *
 * Implementation Note: See aioIntEnable() enable/disable mechanism.
 */

STATUS aioIntDisable (
		      AIO_DD * dev)
    {
    DBG_ENTER;

    if (isEntryError (dev))
	return (ERROR);

    aioDoRegisterAction (dev, dev->IRQ_Enable, "IRQ_Enable");

    dev->intEnabled = FALSE;

    aioDoRegisterAction (dev, dev->IRQ_Clear, "IRQ_Clear");

    DBG_EXIT_OK;
    return (OK);
    }

/*******************************************************************************
 *
 * aioIntEnable - enable interrupts from the card
 *
 * This routine enables Change-of-State (CoS) interrupts from the card. A CoS
 * interrupt occurs whenever any input pin changes state due to an external
 * event. Calling this function when interrupts are enabled is not an error.
 * Interrupts are not generated for ports set for output.
 *
 * RETURNS: OK or ERROR per errno.
 *
 * ERRNO
 *   ENXIO       Driver not installed.
 *   ENODEV      Card is not open.
 *   EACCES      Card is open for another task.
 *
 * Implementation Note
 * -------------------
 * To get Change of State (CoS) interrupts requires five steps plus one:
 * (1) Call vxbIntConnect(). Done in our Attach function.
 * (2) Call vxbIntEnable(). Done in our Attach function. Could toggle this and
 *     vxbIntDisable() to control interrupts, but do not--as far as VxWorks is
 *     concerned, the interrupts are permanently enabled. Thus, vxbIsrVecsShow()
 *     will always show [CONNECTED] and [ENABLED] for our DIO interrupt; use
 *     aioShow() in this driver to show whether CoS interrupts are enabled.
 * (3) At the card level, write 0xf8 to REG_COS_MASK. Could be done at the end 
 *     of Attach (but not before calling iosDevAdd(), but it is best to leave 
 *     the REG_COS_MASK register in its initial state if interrupts are not used.
 * (4) The user must call aioConfigure() with argument PORT_DIR_INPUT for
 *     each port on which interrupts are wanted. Note: This is the reset-default 
 *     state of each port, thus is only necessary if they've been set to output.
 * (5) At the card level, write 0xff to REG_IRQ_EN--the purpose of this function;
 *     writing 0x00 disables the CoS interrupts. Use this mechanism here and in
 *     aioIntDisable() to control whether interrupts are enabled or not.
 *
 * (6) The user must call aioIntCallbackInstall(); while this step is not
 *     required, if the five steps above are done, the interrupt will occur, the
 *     ISR will be invoked, and interrupt cleared--but there will be no effect
 *     on user tasks beyond slight delays.
 */

STATUS aioIntEnable (
		     AIO_DD * dev)
    {
    DBG_ENTER;

    if (isEntryError (dev))
	return (ERROR);

    aioDoRegisterAction (dev, dev->IRQ_Enable, "IRQ_Enable");
    dev->intEnabled = TRUE;

    DBG_EXIT_OK;
    return (OK);
    }

/*******************************************************************************
 *
 * aioOpen - open a card and assign it to the user
 *
 * A card must be opened before it can be used. 
 * This function opens a card and assigns it to the calling user task.
 * It stays open for the exclusive use of the calling user until it is freed by 
 * calling aioClose() or aioCloseForce();
 *
 * It is an error to call this function if the card is already opened for either
 * the current task or another task.
 *
 * RETURNS: OK or ERROR per errno.
 *
 * ERRNO
 *   ENXIO       Driver not installed.
 *   ENODEV      Card is already open for this task.
 *   EACCES      Card is open for another task.
 */

AIO_DD * aioOpen (
		  UINT32 dev)
    {
    DBG_ENTER;
    if (dev > AIO_LastDevNum)
	{
	EXIT_ERR(ENODEV);
	return (NULL);
	}

    if (aioDD[dev].taskId == 0)
	{
	aioDD[dev].taskId = taskIdSelf ();
	DBG_EXIT_OK;
	return (&aioDD[dev]);
	}

    /* Card already open. Set errno depending on whether this task has it or
     * another task does. */

    if (aioDD[dev].taskId == taskIdSelf ())
	EXIT_ERR(ENODEV);
    else
	EXIT_ERR(EACCES);

    return (NULL);
    }

/*******************************************************************************
 *
 * aioPciAttach - attach the ACCES DIO card during PCI bus enumeration
 *
 * Attach an ACCES card. Our device matched its probe so this Attach
 * function is called. Its job is to add the *device* to the I/O sub-system.
 * After this function completes successfully, the device is ready to be opened,
 * read/written, etc. (by calls to our device-specific aioOpen(), etc.).
 * --J2H TO DO: add support for "open" and "create"?
 */

/*
 * To add the AIO device to the I/O subsystem, this function:
 *
 * - Points our module LOCAL paioDD at our module LOCAL device descriptor
 *   aioDD struct (it's initial values will be all zero--sufficient for this
 *   case). (NOTE: A VxBus driver will often allocate the device descriptor
 *   dynamically for each instance of a kind of device, perhaps using a "create"
 *   function for the type of device (e.g., a serial terminal). Because this
 *   version supports only one AIO card in the system and uses a static device
 *   descriptor, no "create" function is needed, but because this pointer
 *   is to the *device* descriptor, it is more correct to set it in this device
 *   Attach function than in the aioPciDrvInit() function).
 *
 * - Uses an "address resource" to get the virtual address of our device (it's
 *   "bar", which maps to the devices physical Base Address Register, BAR) and
 *   a related "handle" (accessing a device register requires the handle, the
 *   bar, and the offset of the register from the bar) and stores the bar, the
 *   handle, and the pointer to the address resource in our device descriptor.
 *   See pAdrRes under topic "DIO Device Descriptor and Pointer to It" above
 *   for details.
 *
 * - Gets and initializes an "interrupt resource" for interrupts, then connects
 *   and enables them. See pIntRes under topic "DIO Device Descriptor and Pointer
 *   to It" above for details.
 *
 * - Saves a pointer to product name from the configuration data in our device
 *   descriptor.
 *
 * - Finally, call iosDevAdd() to add our DIO device to the I/O subsystem.
 *   (Calling it at the end avoids having to delete the device on any failure
 *   of the above steps.)
 *
 * RETURNS: OK or ERROR per ERRNO.
 *
 * ERRNO
 *   ENODEV      Driver is already installed--invalid to call this function.
 *   ENOMEM      Resource allocation failed (BAR address or interrupt).
 *   ENXIO       Driver init error (not done or Probe error) or failure in
 *               this function other than the above (e.g., cannot set up
 *               interrupts).
 *   EIO         Problem with BAR or IRQ (e.g., BAR is invalid 0).
 */

LOCAL STATUS aioPciAttach (
			   struct vxbDev * pVxbDev) /* VxBus device struct, NOT our device descriptor */
    {
    AIO_PCI_DEV_CFG * paioPciDevCfg;
    VXB_RESOURCE_ADR * pAdrRes_pRes;
    DEV_HDR * pDevHdr; /* For convenience. See below. */

#ifdef ACCES_DEBUG
    aio_pVxbDev = pVxbDev;
#endif /* ACCES_DEBUG */

    DBG_MSG (3, "Enter. DeviceNumber = %d\n", AIO_LastDevNum);

    /* Install the driver with all NULL arguments as explained above. */
    drvNum = iosDrvInstall (NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    if (drvNum == ERROR)
	{
	DBG_MSG (1, "iosDrvInstall() call to install driver failed.\n");
	return (ERROR);
	}
    else
	DBG_MSG (2, "iosDrvInstall() OK\n");

    /* Point the paioDD pointer used throughout this driver to our module
     * LOCAL device descriptor at our module LOCAL device descriptor. For
     * convenience, point pDevHdr at its first member, the DEV_HDR.
     */
    AIO_DD *paioDD = &aioDD[AIO_LastDevNum + 1];

    pDevHdr = (DEV_HDR *) paioDD;

    for (UINT16 i = 0; i < 6; i++)
	{
	DBG_MSG (3, "trying alloc IO BAR[%d]\n", i);

	paioDD->pAdrRes = vxbResourceAlloc (pVxbDev, VXB_RES_IO, i);
	/* if BAR[i] is not "I/O", try "MEM" */
	if (paioDD->pAdrRes == NULL)/* if BAR[i] is not "I/O", try "MEM" */
	    {
	    DBG_MSG (3, "IO BAR failed, trying MEM BAR[%d]\n", i );
	    paioDD->pAdrRes = vxbResourceAlloc (pVxbDev, VXB_RES_MEMORY, i);
	    if (paioDD->pAdrRes == NULL)
		{
		DBG_MSG (3, "MEM BAR[%d] failed\n", i);
		continue;
		/* not all BAR[] have resources allocated */
		}
	    }

	pAdrRes_pRes = (VXB_RESOURCE_ADR *) paioDD->pAdrRes->pRes;
	if (pAdrRes_pRes == NULL)
	    {
	    DBG_MSG (1, "pRes is NULL, exiting %s.\n", __func__);
	    EXIT_ERR(ENOMEM);
	    goto cleanup;
	    }

	paioDD->handle[i] = pAdrRes_pRes->pHandle;
	paioDD->bar[i] = (ULONG) pAdrRes_pRes->virtual;

	DBG_MSG (2, "got bar 0x%lx for BAR[%d]\n", paioDD->bar[i], i);

	} // TO DO ?: check for __no__ successes allocating BARs

    /*
     * Get and initialize a resource for the Change of State interrupts, and
     * connect and enable them at the VxWorks driver level and partly at the
     * card level (see aioIntEnable() for more).
     *
     * The third argument to vxbIntConnect() is our ISR, and fourth argument
     * is a pointer to an argument to be passed to our ISR, aioInt().
     *
     * Three more steps are required after this to get interrupts. See
     * aioIntEnable() for details.
     */

    paioDD->pIntRes = vxbResourceAlloc (pVxbDev, VXB_RES_IRQ, 0);
    if (paioDD->pIntRes == NULL)
	{
	EXIT_ERR(ENOMEM);
	goto cleanup;
	}

    DBG_MSG (3, "alloc pIntRes = VXB_RES_IRQ =%p\n", paioDD->pIntRes);

    if (vxbIntConnect (pVxbDev, paioDD->pIntRes, (VOIDFUNCPTR) aioInt,
		       paioDD) != OK)
	{
	EXIT_ERR(ENXIO);
	goto cleanup;
	}

    DBG_MSG (3, "vxbIntConnect() OK.\n");

    if (vxbIntEnable (pVxbDev, paioDD->pIntRes) != OK)
	{
	EXIT_ERR(ENXIO);
	goto cleanup;
	}

    DBG_MSG (3, "vxbIntEnable() OK.\n");

    // Save data from the configuration data table into our device's descriptor.

    paioPciDevCfg = (AIO_PCI_DEV_CFG *) vxbDevDrvDataGet (pVxbDev);
    if (paioPciDevCfg == NULL || paioPciDevCfg->Model == NULL)
	{
	EXIT_ERR(ENXIO);
	goto cleanup;
	}

    paioDD->pProductName = paioPciDevCfg->Model;
    paioDD->IRQ_Init = paioPciDevCfg->IRQ->Init;
    paioDD->IRQ_Disable = paioPciDevCfg->IRQ->Disable;
    paioDD->IRQ_Clear = paioPciDevCfg->IRQ->Clear;
    paioDD->IRQ_Enable = paioPciDevCfg->IRQ->Enable;
    paioDD->IRQ_Generate = paioPciDevCfg->IRQ->Generate;
    paioDD->IRQ_IsLatched = paioPciDevCfg->IRQ->IsLatched;

    // TODO: add get/set functions for ~all TRegisterActions so customers can override IRQ_Enable, for example

    /* Finally, call iosDevAdd() to add our DIO device to the I/O subsystem.
     * Arguments:
     * - Pointer to the DEV_HDR for the device (local pDevHdr, set above--
     *   the first member of our aioDD device descriptor).
     * - Hard-coded device name "/aio/0". When extended to multiple
     *   DIO cards, a local static variable in this file could be initialized
     *   in aioPciDrvInit() and incremented for each device as it is
     *   attached.
     * - The driver number returned by iosDrvInstall() called from our
     *   aioPciAttach() function and stored in module-level static
     *   drvNum.
     */
    ++AIO_LastDevNum;
    char device[12];
    sprintf (device, "/aio/%d", AIO_LastDevNum);

    if (iosDevAdd (pDevHdr, device, drvNum) == ERROR)
	{
	EXIT_ERR(ENXIO);
	goto cleanup;
	}

    DBG_MSG (2, "iosDevAdd() OK. device name=%s device # %d\n", device, AIO_LastDevNum);

    aioDoRegisterAction (paioDD, paioDD->IRQ_Init, "IRQ_Init");

    DBG_EXIT_OK;
    return (OK);

    /*
     * Common exit for errors that require freeing memory allocated here. errno
     * should be set and the "Exit ERROR" message already issued before jumping
     * here.
     */

    cleanup: if (paioDD->pAdrRes != NULL)
	(void) vxbResourceFree (pVxbDev, paioDD->pAdrRes);

    if (paioDD->pIntRes != NULL)
	{
	(void) vxbResourceFree (pVxbDev, paioDD->pIntRes);
	}
    EXIT_ERR(ERROR);
    return (ERROR);
    }

/*******************************************************************************
 *
 * aioPciDrvInit - initialize the ACCES DIO driver
 *
 * This function serves three purposes:
 *
 * 1. It provides a placeholder for partial initialization of this driver. This
 *    function is called during VxWorks initialization by VIP:prjConfigs.c:
 *    usrRoot()) when component DRV_ACCES_DIO_PCI is configured into VxWorks
 *    because of the INIT_RTN and _INIT_ORDER statements in our component
 *    description file 40vxbaio.cfd.
 *
 *    A driver initialization function might call iosDrvInstall() to register the
 *    driver with the I/O system, might set an initial unit number, might take
 *    other actions to support multiple instances of a devices. However, this
 *    driver supports only one DIO card in a system initially, and this function
 *    serves mainly purposes 2 and 3 below. (Calling iosDrvInstall() is deferred
 *    to our Attach function to avoid complications related to when this function
 *    is called--see Attach for details).
 *
 * 2. Because it is called during system initialization, the linker includes the
 *    file in the image. If no function in the file is called, then the file is
 *    not included in the image.
 *
 * 3. Sets up for debug output if ACCES_DEBUG is defined.
 *
 * RETURNS: N/A
 *
 * ERRNO N/A
 */
void aioPciDrvInit (
		    void)
    {
#ifdef ACCES_DEBUG
    /* Set up for DBG_... statements by initializing the kwrite function to use
     * kwriteVga. */
    // _func_kwrite = kwriteVga;
#endif /* ACCES_DEBUG */

    DBG_ENTER;DBG_EXIT_OK;DBG_ENTER;

    DBG_EXIT_OK;
    }

/*******************************************************************************
 *
 * aioPciProbe - probe for device presence
 *
 * Respond to a probe from higher-level VxBus setup code for a device discovered
 * on a PCI bus. Look in our "match" table for an entry with matching vendor and
 * device ID. The argument to this function is a pointer to a vxbDev structure
 * (to avoid confusion--not our device descriptor aioDD struct). It is
 * passed from this Probe function to our Attach by the VxBus device enumeration
 * process, and we can use a member set by calling vxbDevDrvDataSet() to pass a
 * pointer to our configuration data.
 *
 * RETURNS: OK if a matching device is found, else ERROR. (VxBus supports a "best
 * match" feature--see vxbLib.c:vxbDevProbe(); it is not used here.)
 *
 * ERRNO: N/A
 */

LOCAL STATUS aioPciProbe (
			  struct vxbDev * pVxbDev /* VxBus device struct, NOT our device descriptor */
			  )
    {
#ifdef ACCES_DEBUG
    /* Set up for DBG_... statements by initializing the kwrite function to use
     * kwriteVga. */
    _func_kwrite = kwriteVga;  // --J2H debug
#endif /* ACCES_DEBUG */

    const AIO_PCI_DEV_CFG *pMatch = NULL;

#if 0 // old method using vxb match function        
    /* In this case, ERROR means "not a match"; don't set errno. */
    if (vxbPciDevMatch (pVxbDev, matchTable, &pMatch) != OK)
	{
	return (ERROR);
	}
#else // new method using self-matching technique

    PCI_HARDWARE *pPciDevInfo = (PCI_HARDWARE *) vxbDevIvarsGet (pVxbDev);

    if (pPciDevInfo->pciVendId != AIO_VENDORID)
	{
	DBG_MSG0(4, " not us, exiting probe\n");
	return (ERROR); // if this board is not ACCES VENDOR ID, reject probe -- fast!
	}

    (*_func_kprintf) ("\nVenDev: %04X:%04X", pPciDevInfo->pciVendId,
		      pPciDevInfo->pciDevId);
    int i = 0;
    while (matchTable[i].pciDevId != 0x0000)
	{
	if (pPciDevInfo->pciDevId == matchTable[i].pciDevId)
	    {
	    pMatch = &matchTable[i];
	    DBG_MSG0(2, " us, exiting probe\n");
	    break;
	    }
	i++;
	}
    if (pMatch == NULL)
	{
	DBG_MSG0(1, " not a known DeviceID, exiting probe\n");
	return (ERROR); // yes our vendor, but not known card in list
	}
#endif

    /* Match! Save a pointer to the Match Table entry in the vxbDev argument for
     * our Attach function. */

    vxbDevDrvDataSet (pVxbDev, (void *) pMatch);
    DBG_EXIT_OK;
    return (OK);
    }

/*
 * aioShowRegisterAction - for use from non-IRQ contexts
 */
void aioShowRegisterAction (
			    TRegAct act,
			    char * desc)
    {
    printf (eActionFmt[act.Action], desc, eActionNames[act.Action], act.BAR,
	    act.offset, act.value, act.mask);
    }

/******************************************************************************
 *
 * aioShow - show driver and device details for the DIO card
 *
 * Show the current driver and device details for selected CardNumbers.
 * Pass "-1" for to print data for every device, or pass CardNumber to show a 
 * single device's information.
 *  
 * If the driver is not initialized or the card
 * is not attached, the function reports that but does not return ERROR.
 *
 * For output, uses printf().
 *
 * RETURNS: N/A
 *
 * ERRNO: N/A
 */
void aioShow (
	      int CardNumber)
    {

    int first, last;
    if (CardNumber == -1)
	{
	first = 0;
	last = AIO_LastDevNum;
	}
    else
	{
	first = CardNumber;
	last = CardNumber;
	}

    printf ("\nACCES Driver and Device Status\n");
    printf ("--------------------------------------------------------------\n");
    aioGetDriverRevision ();
    for (int cardnum = first; cardnum <= last; cardnum++)
	{
	DEV_HDR * pDevHdr = (DEV_HDR *) &aioDD[cardnum];
	printf ("Card[%2d] ", cardnum);
	if ((pDevHdr == NULL) || (aioDD[cardnum].pProductName == NULL))
	    {
	    printf (
	    " AIO driver not installed or Probe() or Attach() failed.\n");
	    printf (" (device descriptor (aioDD) @ %p)\n", &aioDD[cardnum]);
	    return;
	    }
	printf ("dev=\"%s\"", pDevHdr->name);
	printf (" \"%s\"\n", aioDD[cardnum].pProductName);
	for (int i = 0; i < 6; i++)
	    {
	    if (0 != aioDD[cardnum].bar[i])
		printf ("  BAR[%d]=0x%lx\n", i, aioDD[cardnum].bar[i]);
	    }

	aioShowRegisterAction (aioDD[cardnum].IRQ_Init, 	"IRQ_Init     ");
	aioShowRegisterAction (aioDD[cardnum].IRQ_IsLatched, 	"IRQ_IsLatched");
	aioShowRegisterAction (aioDD[cardnum].IRQ_Disable, 	"IRQ_Disable  ");
	aioShowRegisterAction (aioDD[cardnum].IRQ_Clear, 	"IRQ_Clear    ");
	aioShowRegisterAction (aioDD[cardnum].IRQ_Enable, 	"IRQ_Enable   ");

	// printf ("    interrupt callback funcptr=%p\n", &aioDD[cardnum].pIntCallback);
	//printf ("    intCount=%u\n", &aioDD[cardnum].intCount);
	}
    printf ("%d ACCES I/O Products, Inc. device%s found.\n", AIO_LastDevNum + 1,
	    ((AIO_LastDevNum + 1) == 1) ? "" : "s");
    }

/*******************************************************************************
 *
 * isDriverOK - test whether the driver is operating correctly
 *
 * Returns TRUE if the driver appears to be installed and set up, else FALSE.
 *
 * NOTE: Isolated in this function to allow for more complex tests of the driver
 * in later versions.
 *
 * RETURNS: TRUE if the driver appears to be installed and operating correctly,
 * else FALSE.
 */
LOCAL BOOL isDriverOK (
		       AIO_DD * dev)
    {
    return ((dev != NULL) && (drvNum != -1) && (dev->pProductName != NULL));
    }

/*******************************************************************************
 *
 * isEntryError - return TRUE if if any of several error checks fail
 *
 * Return TRUE if any of several checks on our LOCAL DIO device descriptor fail.
 * Intended for use on entry to most user-callable functions in this file.
 *
 * Side effect: Sets errno based on the failing check per the below.
 *
 * RETURNS: TRUE if an error check fails; FALSE if there is no error. Thus, the
 * typical first statement of a user-callable function in this file is:
 *
 *   if (isEntryError (dev)) return ERROR;
 *
 * ERRNO
 *   ENXIO       Driver not installed.
 *   ENODEV      Card is not open.
 *   EACCES      Card is open for another task.
 */
LOCAL BOOL isEntryError (
			 AIO_DD * dev)
    {
    /* Test for driver: likely init, probe, or attach error. */
    if (!isDriverOK (dev))
	{
	EXIT_ERR(ENXIO);
	return (TRUE);
	}

    /* Test for card not open. */
    if (dev->taskId == 0)
	{
	EXIT_ERR(ENODEV);
	return (TRUE);
	}

    /* Test for card open for another task. */
    if (dev->taskId != taskIdSelf ())
	{
	EXIT_ERR(EACCES);
	return (TRUE);
	}

    return (FALSE);
    }




UINT32 RelInPortB (
		   UINT32 CardNumber,
		   UINT8 reg)
    {
    if (isEntryError (&aioDD[CardNumber]))
	return AIO_Rel_Error_Sentinel;
    return (vxbRead8 (aioDD[CardNumber].handle[2],
		      (UINT8 *) aioDD[CardNumber].bar[2] + reg));
    }

UINT32 RelInPortW (
		   UINT32 CardNumber,
		   UINT8 reg)
    {
    if (isEntryError (&aioDD[CardNumber]))
	return AIO_Rel_Error_Sentinel;
    return (vxbRead16 (aioDD[CardNumber].handle[2],
		       (UINT16 *) aioDD[CardNumber].bar[2] + reg));
    }

UINT32 RelInPortL (
		   UINT32 CardNumber,
		   UINT8 reg)
    {
    if (isEntryError (&aioDD[CardNumber]))
	return AIO_Rel_Error_Sentinel;
    return (vxbRead32 (aioDD[CardNumber].handle[2],
		       (UINT32 *) aioDD[CardNumber].bar[2] + reg));
    }

UINT32 RelOutPortB (
		    UINT32 CardNumber,
		    UINT8 reg,
		    UINT8 value)
    {
    if (isEntryError (&aioDD[CardNumber]))
	return AIO_Rel_Error_Sentinel;
    vxbWrite8 (aioDD[CardNumber].handle[2],
	       (UINT8 *) aioDD[CardNumber].bar[2] + reg, value);
    return (value);
    }

UINT32 RelOutPortW (
		    UINT32 CardNumber,
		    UINT8 reg,
		    UINT16 value)
    {
    if (isEntryError (&aioDD[CardNumber]))
	return AIO_Rel_Error_Sentinel;
    vxbWrite16 (aioDD[CardNumber].handle[2],
		(UINT16 *) aioDD[CardNumber].bar[2] + reg, value);
    return (value);
    }

UINT32 RelOutPortL (
		    UINT32 CardNumber,
		    UINT8 reg,
		    UINT32 value)
    {
    if (isEntryError (&aioDD[CardNumber]))
	return AIO_Rel_Error_Sentinel;
    vxbWrite32 (aioDD[CardNumber].handle[2],
		(UINT32 *) aioDD[CardNumber].bar[2] + reg, value);
    return (value);
    }

/*
 * doRegisterAction()
 * 
 * Accepts a device pointer and Taction, then does one with the other.
 * 
 * This function is currently implemented as an enum-based switch ... mostly because I 
 * don't know how I'd implement it with function pointers and params, nor if it would 
 * execute faster if I did.
 * 
 * prints a description; I *really* want default params grr (J2H) ...desc=" ");
 *
 */
UINT32 aioDoRegisterAction (
			 AIO_DD * dev,
			 TRegAct act,
			 char * desc)
    {
    DBG_showRegisterAction(3, act, desc);

    switch (act.Action)
	{
	case READ8:
	    act.value = vxbRead8 (
	    dev->handle[act.BAR],
	    (UINT8 *) dev->bar[act.BAR] + (UINT8) act.offset);
	    break;
	case READ16:
	    act.value = vxbRead16 (
	    dev->handle[act.BAR],
	    (UINT16 *) dev->bar[act.BAR] + (UINT16) act.offset);
	    break;
	case READ32:
	    act.value = vxbRead32 (dev->handle[act.BAR],
				   (UINT32 *) dev->bar[act.BAR] + act.offset);
	    break;

	case TESTBITS8: // if masked read matches masked value return TRUE
	    act.value = !((vxbRead8 (dev->handle[act.BAR],
				     (UINT8 *) dev->bar[act.BAR] + act.offset)
	    ^ act.value) & act.mask);
	    break;
	case TESTBITS16:
	    act.value = !((vxbRead16 (dev->handle[act.BAR],
				      (UINT16 *) dev->bar[act.BAR] + act.offset)
	    ^ act.value) & act.mask);
	    break;
	case TESTBITS32:
	    act.value = !((vxbRead32 (dev->handle[act.BAR],
				      (UINT32 *) dev->bar[act.BAR] + act.offset)
	    ^ act.value) & act.mask);
	    break;

	case WRITE8:
	    vxbWrite8 (dev->handle[act.BAR],
		       (UINT8 *) dev->bar[act.BAR] + act.offset,
		       (UINT8) act.value);
	    break;
	case WRITE16:
	    vxbWrite16 (dev->handle[act.BAR],
			(UINT16 *) dev->bar[act.BAR] + act.offset,
			(UINT16) act.value);
	    break;
	case WRITE32:
	    vxbWrite32 (dev->handle[act.BAR],
			(UINT32 *) dev->bar[act.BAR] + act.offset,
			(UINT32) act.value);
	    break;
	case WRITE48:
	    //todo:	vxbWrite48( dev->handle[act.BAR], (UINT32 *)dev->bar[act.BAR] + act.offset, (UINT32) act.value );
	    break;

	case TOGGLEBITS8:
	    vxbWrite8 (
	    dev->handle[act.BAR],
	    (UINT8 *) dev->bar[act.BAR] + act.offset,
	    vxbRead8 (dev->handle[act.BAR],
		      (UINT8 *) dev->bar[act.BAR] + act.offset)
	    ^ (UINT8) act.mask);
	    break;
	case TOGGLEBITS16:
	    vxbWrite16 (
	    dev->handle[act.BAR],
	    (UINT16 *) dev->bar[act.BAR] + act.offset,
	    vxbRead16 (dev->handle[act.BAR],
		       (UINT16 *) dev->bar[act.BAR] + act.offset)
	    ^ (UINT16) act.mask);
	    break;
	case TOGGLEBITS32:
	    vxbWrite32 (
	    dev->handle[act.BAR],
	    (UINT32 *) dev->bar[act.BAR] + act.offset,
	    vxbRead32 (dev->handle[act.BAR],
		       (UINT32 *) dev->bar[act.BAR] + act.offset) ^ act.mask);
	    break;

	case WRITEBITS8:
	    vxbWrite8 (
	    dev->handle[act.BAR],
	    (UINT8 *) dev->bar[act.BAR] + act.offset,
	    (vxbRead8 (dev->handle[act.BAR],
		       (UINT8 *) dev->bar[act.BAR] + act.offset) & act.mask)
	    | (UINT8) act.value);
	    break;
	case WRITEBITS16:
	    vxbWrite16 (
	    dev->handle[act.BAR],
	    (UINT16 *) dev->bar[act.BAR] + act.offset,
	    (vxbRead16 (dev->handle[act.BAR],
			(UINT16 *) dev->bar[act.BAR] + act.offset) & act.mask)
	    | (UINT16) act.value);
	    break;
	case WRITEBITS32:
	    vxbWrite32 (
	    dev->handle[act.BAR],
	    (UINT32 *) dev->bar[act.BAR] + act.offset,
	    (vxbRead32 (dev->handle[act.BAR],
			(UINT32 *) dev->bar[act.BAR] + act.offset) & act.mask)
	    | act.value);
	    break;

	case NOT_YET_IMPLEMENTED:
	    break;
	case ERROR_PARSING:
	    break; // shrug. it was intended for when we're loading these scripts off a json/xml/???
	case NotApplicable:
	    break; // "do nothing"
	    break;
	}
    return (act.value);
    }
