/* vxbAIOWDM.h - ACCES I/O Products driver header file*/

/* Copyright (c) 2019, ACCES I/O Products, Inc.
 * 10623 Roselle Street San Diego CA 92121-1506 USA
 * +1-858-550-9559 | +1-800-326-1649 | sales@accesio.com
 *
 * The right to copy, distribute, modify, or otherwise make use of this software
 * may be pursuant to the terms of an applicable ACCES I/O Products, Inc.
 * (ACCES) license agreement, or in the absence of such agreement, is limited to
 * use with ACCES hardware products.
 */

/*
modification history
--------------------
17apr19, aio    First release.
*/

/*
DESCRIPTION
===========
This header file defines and documents the user interface for developers using
various ACCES I/O Products, Inc. I/O cards, on a PCI-type bus, including 
PCI Express, PCI Express Mini Card, and all other compatible busses.

This version of the driver supports:

- A single ACCES card in the system, with device name /aio/0. To access
  the device, a user task must call aioOpen() to reserve it, giving that
  task exclusive access until a call to aioClose(). Thus, no special
  considerations are required to avoid conflict among multiple tasks or CPUs.

- Kernel mode only; no RTP support.
*/

#ifndef __INCaioDrvH
#define __INCaioDrvH

#ifdef __cplusplus
extern "C" {
#endif

#include <vxWorks.h>

/*******************************************************************************
* Types and Defines
* =================
*/
#define AIO_MAX_DEVICES 8
/*
 * Taction is the type of register operation to be performed, used by doRegAct (doRegisterAction)
 */

typedef enum 
	{
		NotApplicable = 0,							// do nothing
		READ8, READ16, READ32,						// read a register
		WRITE8, WRITE16, WRITE32, WRITE48,			// write a register
		TESTBITS8, TESTBITS16, TESTBITS32,			// reads a register and returns true if the masked bits pattern matches
		WRITEBITS8,	WRITEBITS16, WRITEBITS32,		// modifies some bits in a register (usually via read-modify-write operation)
		TOGGLEBITS8, TOGGLEBITS16, TOGGLEBITS32,	// toggle a bit (or some bits) in a register (via read-modify-write, usually)
		NOT_YET_IMPLEMENTED, ERROR_PARSING, 		// someday we want to load these scripts out of an XML/JSON file or whatnot
	} Taction;
LOCAL const char * eActionNames[] = 
					{"Not Applicable",
					 "READ8","READ16","READ32",
					 "WRITE8","WRITE16","WRITE32","WRITE48",
					 "TESTBITS8","TESTBITS16","TESTBITS32",
					 "WRITEBITS8","WRITEBITS16","WRITEBITS32",
					 "TOGGLEBITS8","TOGGLEBITS16","TOGGLEBITS32",
					 "NYI","ERROR_PARSING",
					};

LOCAL const char * eActionFmt[]=
					{"[%15s]\n",
					 "[%15s @bar[%02hx]+%02x]\n","[%15s @bar[%02hx]+%02x]\n","[%15s @bar[%02hx]+%02x]\n",
					 "[%15s @bar[%02hx]+%02x, .v=%04lx]\n","[%15s @bar[%02hx]+%02x, .v=%04lx]\n", "[%15s @bar[%02hx]+%02x, .v=%04lx]\n", "[%15s @bar[%02hx]+%02x, .v=%04lx]\n",
					 "[%15s @bar[%02hx]+%02x, .v=%04lx, .m=%04lx]\n","[%15s @bar[%02hx]+%02x, .v=%04lx, .m=%04lx]\n","[%15s @bar[%02hx]+%02x, .v=%04lx, .m=%04lx]\n",
					 "[%15s @bar[%02hx]+%02x, .v=%04lx, .m=%04lx]\n","[%15s @bar[%02hx]+%02x, .v=%04lx, .m=%04lx]\n","[%15s @bar[%02hx]+%02x, .v=%04lx, .m=%04lx]\n",
					 "[%15s]\n","[%15s @bar[%02hx]+%02x, .v=%04lx, .m=%04lx]\n",
					};

typedef struct _TRegAct TRegAct;	

typedef void (*aioActionHandler)(TRegAct *action);

typedef struct _TRegAct
{
    Taction Action;
    UINT8  BAR;
    UINT16 offset; 					// register index
    UINT32 value;
    UINT32 mask;					// which bits in Value should be asserted (for WRITEBITS) or tested (for TESTBITS)
    aioActionHandler func;	//optional function to perform action
} TRegAct; // TRegisterAction

typedef struct AIO_DD AIO_DD;

#define AIO_Rel_Error_Sentinel (0xAA55AA55)

/*******************************************************************************
* User-Callable Function Prototypes
* =================================
* This driver provides the following user-callable functions (presented
* alphabetically in standard VxWorks format). There are no higher-level
* functions such as open() and ioctl().
*/

/*******************************************************************************
*
* aioClose - close the ACCES card and make it available to another user
*
* This function closes the card and makes it available to another task. It is
* not an error to close the card if it is not open. It is an error for a task
* to attempt to close an open the card if it is not the owner (that is, if it is
* not the task that opened it). Closing the card disables interrupts if enabled,
* and always clears any pending interrupts and resets the ISR callback function
* pointer to NULL.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   EACCES      Card is open for another task.
*/
STATUS aioClose (AIO_DD *dev);


/*******************************************************************************
*
* aioCloseForce - close the DIO card even if not the owner
*
* This function closes the card even if the calling task is not the owner. No
* error checks. It is intended for internal use and is not in the header file.
* See aioClose() for the user-callable "safe" version.
*
* RETURNS: N/A
*/
void aioCloseForce (AIO_DD *dev);


/******************************************************************************
*
* aioGetDriverRevision - show driver revision and build date/time
*
* Show the current driver and device details for the single DIO card supported
* in this version of this driver.  Does not require hardware. Does not generate 
* errors.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/
double aioGetDriverRevision (void);


/*******************************************************************************
*
* aioIntCallbackInstall - install the user interrupt service routine (ISR)
*
* Record the callback interrupt service routine (ISR) function pointer argument.
* The prototype for that function must be:
*
*   STATUS <ISR function name> (void)
*
* The device must be open by calling aioOpen() before calling this routine.
*
* The ISR will be called if:
* - This routine has been called to establish the ISR callback function pointer.
* - Interrupts are enabled by calling aioIntEnable().
* - There is a change of state (CoS) on a pin on a port configured for input
*   while interrupts are enabled.
*
* CoS will be ignored if interrupts are not then enabled, and are ignored for
* pins on a port set for output direction.
*
* This routine may be called at any time that the card is open to change the
* callback function pointer. 
* 
* It may be called with NULL function pointer to remove the callback function.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO:
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/
STATUS aioIntCallbackInstall(
	AIO_DD  * dev,
    STATUS  (*callback_funcptr) (void)   /* callback function pointer */
    );


/*******************************************************************************
*
* aioIntClear - clear a pending interrupt, if any
*
* If an interrupt is pending, this function clears it; otherwise there is no
* effect. NOTE: It is unusual for a user task to call this function because
* interrupts are cleared as part of the interrupt service routine.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/

STATUS aioIntClear (AIO_DD * dev);

/*******************************************************************************
*
* aioIntDisable - disable interrupts from the card
*
* This routine disables interrupts from the card. Any pending interrupts are
* cleared. Calling this routine when interrupts are disabled is not an error.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*
* Implementation Note: See aioIntEnable() enable/disable mechanism.
*/

STATUS aioIntDisable (AIO_DD * dev);

/*******************************************************************************
*
* aioIntEnable - enable interrupts from the card
*
* This routine enables Change-of-State (CoS) interrupts from the card. A CoS
* interrupt occurs whenever any input pin changes state due to an external
* event. Calling this function when interrupts are enabled is not an error.
* Interrupts are not generated for ports set for output.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*
*/
STATUS aioIntEnable (AIO_DD * dev);


/*******************************************************************************
*
* aioOpen - open the DIO card and assign it to the user
* TODO: update comment
* This version of this driver supports a single DIO card in the system. This
* function opens that card and assigns it to the calling user task. A DIO card
* must be opened before it can be used. It stays open for the exclusive use of
* the calling user until it is freed by calling acessDioClose().
*
* It is an error to call this function if the card is already opened for either
* the current task or another task.
*
* RETURNS: NULL on error, (opaque) pointer to device instance data on success.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is already open for this task.
*   EACCES      Card is open for another task.
*/
AIO_DD * aioOpen (UINT8 dev);


/******************************************************************************
*
* aioShow - show driver and device details for the DIO card
*
* Show the current driver and device details for the single DIO card supported
* in this version of this driver. If the driver is not initialized or the card
* is not attached, the function reports that but does not return ERROR.
*
* For output, uses printf().
*
* RETURNS: N/A
*
* ERRNO: N/A
*/
void aioShow (int card);


UINT32 RelInPortB(UINT32 CardNumber, UINT8 reg);
UINT32 RelInPortW(UINT32 CardNumber, UINT8 reg);
UINT32 RelInPortL(UINT32 CardNumber, UINT8 reg);
UINT32 RelOutPortB(UINT32 CardNumber, UINT8 reg, UINT8 value);
UINT32 RelOutPortW(UINT32 CardNumber, UINT8 reg, UINT16 value);
UINT32 RelOutPortL(UINT32 CardNumber, UINT8 reg, UINT32 value);

/* the following two lines are convenience #defines to math the Windows API names */
#define RelInPort  RelInPortW
#define RelOutPort RelOutPortW

#ifdef __cplusplus
}
#endif

#endif  /* __INCaccesIdioDrvh */
